@@ -1,5 +1,6 @@
+# -*- coding: utf-8 -*-
 from __future__ import print_function, division, absolute_import, unicode_literals
-from fontTools.misc.py23 import tounicode, round, BytesIO
+from fontTools.misc.py23 import byteord, tounicode, round, unichr, BytesIO
 
 import logging
 import math
@@ -16,11 +17,17 @@ from fontTools.ttLib.tables.O_S_2f_2 import Panose
 from fontTools.ttLib.tables._h_e_a_d import mac_epoch_diff
 from fontTools.ttLib.tables._g_l_y_f import Glyph, USE_MY_METRICS
 from fontTools.misc.arrayTools import unionRect
+from fontTools.misc.fixedTools import otRound
 
 from ufo2ft.fontInfoData import (
     getAttrWithFallback, dateStringToTimeValue, dateStringForNow,
     intListToNum, normalizeStringForPostscript)
-from ufo2ft.util import makeOfficialGlyphOrder
+from ufo2ft.util import (
+    makeOfficialGlyphOrder,
+    makeUnicodeToGlyphNameMapping,
+    calcCodePageRanges,
+)
+
 
 logger = logging.getLogger(__name__)
 
@@ -35,14 +42,14 @@ def _isNonBMP(s):
     return False
 
 
-def _getVerticalOrigin(glyph):
-    height = glyph.height
+def _getVerticalOrigin(font, glyph):
+    font_ascender = font['hhea'].ascent
     if (hasattr(glyph, "verticalOrigin") and
             glyph.verticalOrigin is not None):
         verticalOrigin = glyph.verticalOrigin
     else:
-        verticalOrigin = height
-    return round(verticalOrigin)
+        verticalOrigin = font_ascender
+    return otRound(verticalOrigin)
 
 
 class BaseOutlineCompiler(object):
@@ -73,12 +80,19 @@ class BaseOutlineCompiler(object):
         """
         self.otf = TTFont(sfntVersion=self.sfntVersion)
 
-        self.vertical = False
-        for glyph in self.allGlyphs.values():
-            if (hasattr(glyph, "verticalOrigin") and
-                    glyph.verticalOrigin is not None):
-                self.vertical = True
-                break
+        # only compile vertical metrics tables if vhea metrics a defined
+        vertical_metrics = [
+            "openTypeVheaVertTypoAscender",
+            "openTypeVheaVertTypoDescender",
+            "openTypeVheaVertTypoLineGap",
+            "openTypeVheaCaretSlopeRise",
+            "openTypeVheaCaretSlopeRun",
+            "openTypeVheaCaretOffset",
+        ]
+        self.vertical = all(
+            getAttrWithFallback(self.ufo.info, metric) is not None
+            for metric in vertical_metrics
+        )
 
         # populate basic tables
         self.setupTable_head()
@@ -104,7 +118,7 @@ class BaseOutlineCompiler(object):
         The bounding box of empty glyphs (without contours or components) is
         set to None.
 
-        Float values are rounded to integers using the built-in round().
+        Float values are rounded to integers using fontTools otRound().
 
         **This should not be called externally.** Subclasses
         may override this method to handle the bounds creation
@@ -123,7 +137,7 @@ class BaseOutlineCompiler(object):
             if glyph or glyph.components:
                 bounds = getControlPointBounds(glyph)
                 if bounds:
-                    bounds = BoundingBox(*(round(v) for v in bounds))
+                    bounds = BoundingBox(*(otRound(v) for v in bounds))
             glyphBoxes[glyphName] = bounds
         return glyphBoxes
 
@@ -157,12 +171,7 @@ class BaseOutlineCompiler(object):
         may override this method to handle the mapping creation
         in a different way if desired.
         """
-        mapping = {}
-        for glyphName, glyph in self.allGlyphs.items():
-            unicodes = glyph.unicodes
-            for uni in unicodes:
-                mapping[uni] = glyphName
-        return mapping
+        return makeUnicodeToGlyphNameMapping(self.allGlyphs, self.glyphOrder)
 
     @staticmethod
     def makeMissingRequiredGlyphs(font, glyphSet):
@@ -176,10 +185,10 @@ class BaseOutlineCompiler(object):
         if ".notdef" in glyphSet:
             return
 
-        unitsPerEm = round(getAttrWithFallback(font.info, "unitsPerEm"))
-        ascender = round(getAttrWithFallback(font.info, "ascender"))
-        descender = round(getAttrWithFallback(font.info, "descender"))
-        defaultWidth = round(unitsPerEm * 0.5)
+        unitsPerEm = otRound(getAttrWithFallback(font.info, "unitsPerEm"))
+        ascender = otRound(getAttrWithFallback(font.info, "ascender"))
+        descender = otRound(getAttrWithFallback(font.info, "descender"))
+        defaultWidth = otRound(unitsPerEm * 0.5)
         glyphSet[".notdef"] = StubGlyph(name=".notdef",
                                         width=defaultWidth,
                                         unitsPerEm=unitsPerEm,
@@ -239,7 +248,7 @@ class BaseOutlineCompiler(object):
                 fullFontRevision, head.fontRevision)
 
         # upm
-        head.unitsPerEm = round(getAttrWithFallback(font.info, "unitsPerEm"))
+        head.unitsPerEm = otRound(getAttrWithFallback(font.info, "unitsPerEm"))
 
         # times
         head.created = dateStringToTimeValue(getAttrWithFallback(font.info, "openTypeHeadCreated")) - mac_epoch_diff
@@ -247,10 +256,10 @@ class BaseOutlineCompiler(object):
 
         # bounding box
         xMin, yMin, xMax, yMax = self.fontBoundingBox
-        head.xMin = round(xMin)
-        head.yMin = round(yMin)
-        head.xMax = round(xMax)
-        head.yMax = round(yMax)
+        head.xMin = otRound(xMin)
+        head.yMin = otRound(yMin)
+        head.xMax = otRound(xMax)
+        head.yMax = otRound(yMax)
 
         # style mapping
         styleMapStyleName = getAttrWithFallback(font.info, "styleMapStyleName")
@@ -265,7 +274,7 @@ class BaseOutlineCompiler(object):
 
         # misc
         head.flags = intListToNum(getAttrWithFallback(font.info, "openTypeHeadFlags"), 0, 16)
-        head.lowestRecPPEM = round(getAttrWithFallback(font.info, "openTypeHeadLowestRecPPEM"))
+        head.lowestRecPPEM = otRound(getAttrWithFallback(font.info, "openTypeHeadLowestRecPPEM"))
         head.fontDirectionHint = 2
         head.indexToLocFormat = 0
         head.glyphDataFormat = 0
@@ -419,11 +428,15 @@ class BaseOutlineCompiler(object):
         table creation in a different way if desired.
         """
         self.otf["OS/2"] = os2 = newTable("OS/2")
+        hmtx = self.otf["hmtx"]
         font = self.ufo
         os2.version = 0x0004
         # average glyph width
-        widths = [glyph.width for glyph in self.allGlyphs.values() if glyph.width > 0]
-        os2.xAvgCharWidth = round(sum(widths) / len(widths))
+        widths = [width for width, _ in hmtx.metrics.values() if width > 0]
+        if widths:
+            os2.xAvgCharWidth = otRound(sum(widths) / len(widths))
+        else:
+            os2.xAvgCharWidth = 0
         # weight and width classes
         os2.usWeightClass = getAttrWithFallback(font.info, "openTypeOS2WeightClass")
         os2.usWidthClass = getAttrWithFallback(font.info, "openTypeOS2WidthClass")
@@ -442,45 +455,45 @@ class BaseOutlineCompiler(object):
         v = getAttrWithFallback(font.info, "openTypeOS2SubscriptXSize")
         if v is None:
             v = unitsPerEm * 0.65
-        os2.ySubscriptXSize = round(v)
+        os2.ySubscriptXSize = otRound(v)
         v = getAttrWithFallback(font.info, "openTypeOS2SubscriptYSize")
         if v is None:
             v = unitsPerEm * 0.6
-        os2.ySubscriptYSize = round(v)
+        os2.ySubscriptYSize = otRound(v)
         v = getAttrWithFallback(font.info, "openTypeOS2SubscriptYOffset")
         if v is None:
             v = unitsPerEm * 0.075
-        os2.ySubscriptYOffset = round(v)
+        os2.ySubscriptYOffset = otRound(v)
         v = getAttrWithFallback(font.info, "openTypeOS2SubscriptXOffset")
         if v is None:
             v = adjustOffset(-os2.ySubscriptYOffset, italicAngle)
-        os2.ySubscriptXOffset = round(v)
+        os2.ySubscriptXOffset = otRound(v)
 
         v = getAttrWithFallback(font.info, "openTypeOS2SuperscriptXSize")
         if v is None:
             v = os2.ySubscriptXSize
-        os2.ySuperscriptXSize = round(v)
+        os2.ySuperscriptXSize = otRound(v)
         v = getAttrWithFallback(font.info, "openTypeOS2SuperscriptYSize")
         if v is None:
             v = os2.ySubscriptYSize
-        os2.ySuperscriptYSize = round(v)
+        os2.ySuperscriptYSize = otRound(v)
         v = getAttrWithFallback(font.info, "openTypeOS2SuperscriptYOffset")
         if v is None:
             v = unitsPerEm * 0.35
-        os2.ySuperscriptYOffset = round(v)
+        os2.ySuperscriptYOffset = otRound(v)
         v = getAttrWithFallback(font.info, "openTypeOS2SuperscriptXOffset")
         if v is None:
             v = adjustOffset(os2.ySuperscriptYOffset, italicAngle)
-        os2.ySuperscriptXOffset = round(v)
+        os2.ySuperscriptXOffset = otRound(v)
 
         v = getAttrWithFallback(font.info, "openTypeOS2StrikeoutSize")
         if v is None:
             v = getAttrWithFallback(font.info, "postscriptUnderlineThickness")
-        os2.yStrikeoutSize = round(v)
+        os2.yStrikeoutSize = otRound(v)
         v = getAttrWithFallback(font.info, "openTypeOS2StrikeoutPosition")
         if v is None:
             v = xHeight * 0.6 if xHeight else unitsPerEm * 0.22
-        os2.yStrikeoutPosition = round(v)
+        os2.yStrikeoutPosition = otRound(v)
 
         # family class
         ibmFontClass, ibmFontSubclass = getAttrWithFallback(
@@ -502,26 +515,34 @@ class BaseOutlineCompiler(object):
         os2.panose = panose
         # Unicode ranges
         uniRanges = getAttrWithFallback(font.info, "openTypeOS2UnicodeRanges")
-        os2.ulUnicodeRange1 = intListToNum(uniRanges, 0, 32)
-        os2.ulUnicodeRange2 = intListToNum(uniRanges, 32, 32)
-        os2.ulUnicodeRange3 = intListToNum(uniRanges, 64, 32)
-        os2.ulUnicodeRange4 = intListToNum(uniRanges, 96, 32)
+        if uniRanges is not None:
+            os2.ulUnicodeRange1 = intListToNum(uniRanges, 0, 32)
+            os2.ulUnicodeRange2 = intListToNum(uniRanges, 32, 32)
+            os2.ulUnicodeRange3 = intListToNum(uniRanges, 64, 32)
+            os2.ulUnicodeRange4 = intListToNum(uniRanges, 96, 32)
+        else:
+            os2.recalcUnicodeRanges(self.otf)
+
         # codepage ranges
         codepageRanges = getAttrWithFallback(font.info, "openTypeOS2CodePageRanges")
+        if codepageRanges is None:
+            unicodes = self.unicodeToGlyphNameMapping.keys()
+            codepageRanges = calcCodePageRanges(unicodes)
         os2.ulCodePageRange1 = intListToNum(codepageRanges, 0, 32)
         os2.ulCodePageRange2 = intListToNum(codepageRanges, 32, 32)
+
         # vendor id
         os2.achVendID = tounicode(
             getAttrWithFallback(font.info, "openTypeOS2VendorID"),
             encoding="ascii", errors="ignore")
         # vertical metrics
-        os2.sxHeight = round(getAttrWithFallback(font.info, "xHeight"))
-        os2.sCapHeight = round(getAttrWithFallback(font.info, "capHeight"))
-        os2.sTypoAscender = round(getAttrWithFallback(font.info, "openTypeOS2TypoAscender"))
-        os2.sTypoDescender = round(getAttrWithFallback(font.info, "openTypeOS2TypoDescender"))
-        os2.sTypoLineGap = round(getAttrWithFallback(font.info, "openTypeOS2TypoLineGap"))
-        os2.usWinAscent = round(getAttrWithFallback(font.info, "openTypeOS2WinAscent"))
-        os2.usWinDescent = round(getAttrWithFallback(font.info, "openTypeOS2WinDescent"))
+        os2.sxHeight = otRound(getAttrWithFallback(font.info, "xHeight"))
+        os2.sCapHeight = otRound(getAttrWithFallback(font.info, "capHeight"))
+        os2.sTypoAscender = otRound(getAttrWithFallback(font.info, "openTypeOS2TypoAscender"))
+        os2.sTypoDescender = otRound(getAttrWithFallback(font.info, "openTypeOS2TypoDescender"))
+        os2.sTypoLineGap = otRound(getAttrWithFallback(font.info, "openTypeOS2TypoLineGap"))
+        os2.usWinAscent = otRound(getAttrWithFallback(font.info, "openTypeOS2WinAscent"))
+        os2.usWinDescent = otRound(getAttrWithFallback(font.info, "openTypeOS2WinDescent"))
         # style mapping
         selection = list(getAttrWithFallback(font.info, "openTypeOS2Selection"))
         styleMapStyleName = getAttrWithFallback(font.info, "styleMapStyleName")
@@ -568,13 +589,103 @@ class BaseOutlineCompiler(object):
         self.otf["hmtx"] = hmtx = newTable("hmtx")
         hmtx.metrics = {}
         for glyphName, glyph in self.allGlyphs.items():
-            width = glyph.width
+            width = otRound(glyph.width)
             if width < 0:
                 raise ValueError(
                     "The width should not be negative: '%s'" % (glyphName))
             bounds = self.glyphBoundingBoxes[glyphName]
             left = bounds.xMin if bounds else 0
-            hmtx[glyphName] = (round(width), left)
+            hmtx[glyphName] = (width, left)
+
+    def _setupTable_hhea_or_vhea(self, tag):
+        """
+        Make the hhea table or the vhea table. This assume the hmtx or
+        the vmtx were respectively made first.
+        """
+        if tag == "hhea":
+            isHhea = True
+        else:
+            isHhea = False
+        self.otf[tag] = table = newTable(tag)
+        mtxTable = self.otf[tag[0] + "mtx"]
+        font = self.ufo
+        if isHhea:
+            table.tableVersion = 0x00010000
+        else:
+            table.tableVersion = 0x00011000
+        # Vertical metrics in hhea, horizontal metrics in vhea
+        # and caret info.
+        # The hhea metrics names are formed as:
+        #   "openType" + tag.title() + "Ascender", etc.
+        # While vhea metrics names are formed as:
+        #   "openType" + tag.title() + "VertTypo" + "Ascender", etc.
+        # Caret info names only differ by tag.title().
+        commonPrefix = "openType%s" % tag.title()
+        if isHhea:
+            metricsPrefix = commonPrefix
+        else:
+            metricsPrefix = "openType%sVertTypo" % tag.title()
+        metricsDict = {
+            "ascent": "%sAscender" % metricsPrefix,
+            "descent": "%sDescender" % metricsPrefix,
+            "lineGap": "%sLineGap" % metricsPrefix,
+            "caretSlopeRise": "%sCaretSlopeRise" % commonPrefix,
+            "caretSlopeRun": "%sCaretSlopeRun" % commonPrefix,
+            "caretOffset": "%sCaretOffset" % commonPrefix,
+        }
+        for otfName, ufoName in metricsDict.items():
+            setattr(table, otfName,
+                    otRound(getAttrWithFallback(font.info, ufoName)))
+        # Horizontal metrics in hhea, vertical metrics in vhea
+        advances = []  # width in hhea, height in vhea
+        firstSideBearings = []  # left in hhea, top in vhea
+        secondSideBearings = []  # right in hhea, bottom in vhea
+        extents = []
+        for glyphName in self.allGlyphs:
+            advance, firstSideBearing = mtxTable[glyphName]
+            advances.append(advance)
+            bounds = self.glyphBoundingBoxes[glyphName]
+            if bounds is None:
+                continue
+            if isHhea:
+                boundsAdvance = (bounds.xMax - bounds.xMin)
+                # equation from the hhea spec for calculating xMaxExtent:
+                #   Max(lsb + (xMax - xMin))
+                extent = firstSideBearing + boundsAdvance
+            else:
+                boundsAdvance = (bounds.yMax - bounds.yMin)
+                # equation from the vhea spec for calculating yMaxExtent:
+                #   Max(tsb + (yMax - yMin)).
+                extent = firstSideBearing + boundsAdvance
+            secondSideBearing = advance - firstSideBearing - boundsAdvance
+
+            firstSideBearings.append(firstSideBearing)
+            secondSideBearings.append(secondSideBearing)
+            extents.append(extent)
+        setattr(table,
+                "advance%sMax" % ("Width" if isHhea else "Height"),
+                max(advances) if advances else 0)
+        setattr(table,
+                "min%sSideBearing" % ("Left" if isHhea else "Top"),
+                min(firstSideBearings) if firstSideBearings else 0)
+        setattr(table,
+                "min%sSideBearing" % ("Right" if isHhea else "Bottom"),
+                min(secondSideBearings) if secondSideBearings else 0)
+        setattr(table,
+                "%sMaxExtent" % ("x" if isHhea else "y"),
+                max(extents) if extents else 0)
+        if isHhea:
+            reserved = range(4)
+        else:
+            # vhea.reserved0 is caretOffset for legacy reasons
+            reserved = range(1, 5)
+        for i in reserved:
+            setattr(table, "reserved%i" % i, 0)
+        table.metricDataFormat = 0
+        # glyph count
+        setattr(table,
+                "numberOf%sMetrics" % ("H" if isHhea else "V"),
+                len(self.allGlyphs))
 
     def setupTable_hhea(self):
         """
@@ -584,47 +695,7 @@ class BaseOutlineCompiler(object):
         may override or supplement this method to handle the
         table creation in a different way if desired.
         """
-        self.otf["hhea"] = hhea = newTable("hhea")
-        hmtx = self.otf["hmtx"]
-        font = self.ufo
-        hhea.tableVersion = 0x00010000
-        # vertical metrics
-        hhea.ascent = round(getAttrWithFallback(font.info, "openTypeHheaAscender"))
-        hhea.descent = round(getAttrWithFallback(font.info, "openTypeHheaDescender"))
-        hhea.lineGap = round(getAttrWithFallback(font.info, "openTypeHheaLineGap"))
-        # horizontal metrics
-        widths = []
-        lefts = []
-        rights = []
-        extents = []
-        for glyphName in self.allGlyphs:
-            width, left = hmtx[glyphName]
-            widths.append(width)
-            bounds = self.glyphBoundingBoxes[glyphName]
-            if bounds is None:
-                continue
-            right = width - left - (bounds.xMax - bounds.xMin)
-            lefts.append(left)
-            rights.append(right)
-            # equation from the hhea spec for calculating xMaxExtent:
-            #   Max(lsb + (xMax - xMin))
-            extent = left + (bounds.xMax - bounds.xMin)
-            extents.append(extent)
-        hhea.advanceWidthMax = max(widths)
-        hhea.minLeftSideBearing = min(lefts)
-        hhea.minRightSideBearing = min(rights)
-        hhea.xMaxExtent = max(extents)
-        # misc
-        hhea.caretSlopeRise = getAttrWithFallback(font.info, "openTypeHheaCaretSlopeRise")
-        hhea.caretSlopeRun = getAttrWithFallback(font.info, "openTypeHheaCaretSlopeRun")
-        hhea.caretOffset = round(getAttrWithFallback(font.info, "openTypeHheaCaretOffset"))
-        hhea.reserved0 = 0
-        hhea.reserved1 = 0
-        hhea.reserved2 = 0
-        hhea.reserved3 = 0
-        hhea.metricDataFormat = 0
-        # glyph count
-        hhea.numberOfHMetrics = len(self.allGlyphs)
+        self._setupTable_hhea_or_vhea("hhea")
 
     def setupTable_vmtx(self):
         """
@@ -638,11 +709,14 @@ class BaseOutlineCompiler(object):
         self.otf["vmtx"] = vmtx = newTable("vmtx")
         vmtx.metrics = {}
         for glyphName, glyph in self.allGlyphs.items():
-            height = glyph.height
-            verticalOrigin = _getVerticalOrigin(glyph)
+            height = otRound(glyph.height)
+            if height < 0:
+                raise ValueError(
+                    "The height should not be negative: '%s'" % (glyphName))
+            verticalOrigin = _getVerticalOrigin(self.otf, glyph)
             bounds = self.glyphBoundingBoxes[glyphName]
             top = bounds.yMax if bounds else 0
-            vmtx[glyphName] = (round(height), verticalOrigin - top)
+            vmtx[glyphName] = (height, verticalOrigin - top)
 
     def setupTable_VORG(self):
         """
@@ -657,12 +731,13 @@ class BaseOutlineCompiler(object):
         vorg.minorVersion = 0
         vorg.VOriginRecords = {}
         # Find the most frequent verticalOrigin
-        vorg_count = Counter(_getVerticalOrigin(glyph)
+        vorg_count = Counter(_getVerticalOrigin(self.otf, glyph)
                              for glyph in self.allGlyphs.values())
         vorg.defaultVertOriginY = vorg_count.most_common(1)[0][0]
         if len(vorg_count) > 1:
             for glyphName, glyph in self.allGlyphs.items():
-                vorg.VOriginRecords[glyphName] = _getVerticalOrigin(glyph)
+                vorg.VOriginRecords[glyphName] = _getVerticalOrigin(
+                    self.otf, glyph)
         vorg.numVertOriginYMetrics = len(vorg.VOriginRecords)
 
     def setupTable_vhea(self):
@@ -674,44 +749,7 @@ class BaseOutlineCompiler(object):
         may override or supplement this method to handle the
         table creation in a different way if desired.
         """
-        self.otf["vhea"] = vhea = newTable("vhea")
-        font = self.ufo
-        head = self.otf["head"]
-        vmtx = self.otf["vmtx"]
-        vhea.tableVersion = 0x00011000
-        # horizontal metrics
-        vhea.ascent = round(getAttrWithFallback(font.info, "openTypeVheaVertTypoAscender"))
-        vhea.descent = round(getAttrWithFallback(font.info, "openTypeVheaVertTypoDescender"))
-        vhea.lineGap = round(getAttrWithFallback(font.info, "openTypeVheaVertTypoLineGap"))
-        # vertical metrics
-        heights = []
-        tops = []
-        bottoms = []
-        for glyphName in self.allGlyphs:
-            height, top = vmtx[glyphName]
-            heights.append(height)
-            bounds = self.glyphBoundingBoxes[glyphName]
-            if bounds is None:
-                continue
-            bottom = height - top - (bounds.yMax - bounds.yMin)
-            tops.append(top)
-            bottoms.append(bottom)
-        vhea.advanceHeightMax = max(heights)
-        vhea.minTopSideBearing = max(tops)
-        vhea.minBottomSideBearing = max(bottoms)
-        vhea.yMaxExtent = vhea.minTopSideBearing - (head.yMax - head.yMin)
-        # misc
-        vhea.caretSlopeRise = getAttrWithFallback(font.info, "openTypeVheaCaretSlopeRise")
-        vhea.caretSlopeRun = getAttrWithFallback(font.info, "openTypeVheaCaretSlopeRun")
-        vhea.caretOffset = getAttrWithFallback(font.info, "openTypeVheaCaretOffset")
-        vhea.reserved0 = 0
-        vhea.reserved1 = 0
-        vhea.reserved2 = 0
-        vhea.reserved3 = 0
-        vhea.reserved4 = 0
-        vhea.metricDataFormat = 0
-        # glyph count
-        vhea.numberOfVMetrics = len(self.allGlyphs)
+        self._setupTable_hhea_or_vhea("vhea")
 
     def setupTable_post(self):
         """
@@ -729,11 +767,9 @@ class BaseOutlineCompiler(object):
         post.italicAngle = italicAngle
         # underline
         underlinePosition = getAttrWithFallback(font.info, "postscriptUnderlinePosition")
-        post.underlinePosition = round(underlinePosition)
+        post.underlinePosition = otRound(underlinePosition)
         underlineThickness = getAttrWithFallback(font.info, "postscriptUnderlineThickness")
-        post.underlineThickness = round(underlineThickness)
-        # determine if the font has a fixed width
-        widths = set([glyph.width for glyph in self.allGlyphs.values()])
+        post.underlineThickness = otRound(underlineThickness)
         post.isFixedPitch = getAttrWithFallback(font.info, "postscriptIsFixedPitch")
         # misc
         post.minMemType42 = 0
@@ -785,8 +821,14 @@ class OutlineOTFCompiler(BaseOutlineCompiler):
 
     sfntVersion = "OTTO"
 
-    def __init__(self, font, glyphSet=None, glyphOrder=None,
-                 roundTolerance=None):
+    def __init__(
+        self,
+        font,
+        glyphSet=None,
+        glyphOrder=None,
+        roundTolerance=None,
+        optimizeCFF=True,
+    ):
         if roundTolerance is not None:
             self.roundTolerance = float(roundTolerance)
         else:
@@ -794,6 +836,7 @@ class OutlineOTFCompiler(BaseOutlineCompiler):
             self.roundTolerance = 0.5
         super(OutlineOTFCompiler, self).__init__(
             font, glyphSet=glyphSet, glyphOrder=glyphOrder)
+        self.optimizeCFF = optimizeCFF
 
     def makeGlyphsBoundingBoxes(self):
         """
@@ -814,7 +857,7 @@ class OutlineOTFCompiler(BaseOutlineCompiler):
             return pen.bounds
 
         def toInt(value, else_callback):
-            rounded = round(value)
+            rounded = otRound(value)
             if tolerance >= 0.5 or abs(rounded - value) <= tolerance:
                 return rounded
             else:
@@ -846,16 +889,22 @@ class OutlineOTFCompiler(BaseOutlineCompiler):
         in a different way if desired.
         """
         width = glyph.width
-        # subtract the nominal width
-        postscriptNominalWidthX = getAttrWithFallback(self.ufo.info, "postscriptNominalWidthX")
-        if postscriptNominalWidthX:
-            width = width - postscriptNominalWidthX
-        # round
-        width = round(width)
+        defaultWidth = private.defaultWidthX
+        nominalWidth = private.nominalWidthX
+        if width == defaultWidth:
+            # if width equals the default it can be omitted from charstring
+            width = None
+        else:
+            # subtract the nominal width
+            width -= nominalWidth
+        if width is not None:
+            width = otRound(width)
         pen = T2CharStringPen(width, self.allGlyphs,
                               roundTolerance=self.roundTolerance)
         glyph.draw(pen)
-        charString = pen.getCharString(private, globalSubrs)
+        charString = pen.getCharString(
+            private, globalSubrs, optimize=self.optimizeCFF
+        )
         return charString
 
     def setupTable_maxp(self):
@@ -906,9 +955,9 @@ class OutlineOTFCompiler(BaseOutlineCompiler):
         if trademark:
             trademark = normalizeStringForPostscript(trademark.replace("\u00A9", "Copyright"))
         if trademark != self.ufo.info.trademark:
-            logger.warning("The trademark was normalized for storage in the "
-                           "CFF table and consequently some characters were "
-                           "dropped: '%s'", trademark)
+            logger.info("The trademark was normalized for storage in the "
+                        "CFF table and consequently some characters were "
+                        "dropped: '%s'", trademark)
         if trademark is None:
             trademark = ""
         topDict.Notice = trademark
@@ -916,56 +965,64 @@ class OutlineOTFCompiler(BaseOutlineCompiler):
         if copyright:
             copyright = normalizeStringForPostscript(copyright.replace("\u00A9", "Copyright"))
         if copyright != self.ufo.info.copyright:
-            logger.warning("The copyright was normalized for storage in the "
-                           "CFF table and consequently some characters were "
-                           "dropped: '%s'", copyright)
+            logger.info("The copyright was normalized for storage in the "
+                        "CFF table and consequently some characters were "
+                        "dropped: '%s'", copyright)
         if copyright is None:
             copyright = ""
         topDict.Copyright = copyright
         topDict.FullName = getAttrWithFallback(info, "postscriptFullName")
         topDict.FamilyName = getAttrWithFallback(info, "openTypeNamePreferredFamilyName")
         topDict.Weight = getAttrWithFallback(info, "postscriptWeightName")
-        topDict.FontName = psName
         # populate various numbers
         topDict.isFixedPitch = getAttrWithFallback(info, "postscriptIsFixedPitch")
         topDict.ItalicAngle = getAttrWithFallback(info, "italicAngle")
         underlinePosition = getAttrWithFallback(info, "postscriptUnderlinePosition")
-        topDict.UnderlinePosition = round(underlinePosition)
+        topDict.UnderlinePosition = otRound(underlinePosition)
         underlineThickness = getAttrWithFallback(info, "postscriptUnderlineThickness")
-        topDict.UnderlineThickness = round(underlineThickness)
+        topDict.UnderlineThickness = otRound(underlineThickness)
         # populate font matrix
-        unitsPerEm = round(getAttrWithFallback(info, "unitsPerEm"))
+        unitsPerEm = otRound(getAttrWithFallback(info, "unitsPerEm"))
         topDict.FontMatrix = [1.0 / unitsPerEm, 0, 0, 1.0 / unitsPerEm, 0, 0]
         # populate the width values
-        defaultWidthX = round(getAttrWithFallback(info, "postscriptDefaultWidthX"))
+        if not any(hasattr(info, attr) and getattr(info, attr) is not None
+                   for attr in ("postscriptDefaultWidthX",
+                                "postscriptNominalWidthX")):
+            # no custom values set in fontinfo.plist; compute optimal ones
+            from fontTools.cffLib.width import optimizeWidths
+            hmtx = self.otf['hmtx']
+            widths = [m[0] for m in hmtx.metrics.values()]
+            defaultWidthX, nominalWidthX = optimizeWidths(widths)
+        else:
+            defaultWidthX = otRound(getAttrWithFallback(info, "postscriptDefaultWidthX"))
+            nominalWidthX = otRound(getAttrWithFallback(info, "postscriptNominalWidthX"))
         if defaultWidthX:
             private.rawDict["defaultWidthX"] = defaultWidthX
-        nominalWidthX = round(getAttrWithFallback(info, "postscriptNominalWidthX"))
         if nominalWidthX:
             private.rawDict["nominalWidthX"] = nominalWidthX
         # populate hint data
-        blueFuzz = round(getAttrWithFallback(info, "postscriptBlueFuzz"))
-        blueShift = round(getAttrWithFallback(info, "postscriptBlueShift"))
+        blueFuzz = otRound(getAttrWithFallback(info, "postscriptBlueFuzz"))
+        blueShift = otRound(getAttrWithFallback(info, "postscriptBlueShift"))
         blueScale = getAttrWithFallback(info, "postscriptBlueScale")
         forceBold = getAttrWithFallback(info, "postscriptForceBold")
         blueValues = getAttrWithFallback(info, "postscriptBlueValues")
         if isinstance(blueValues, list):
-            blueValues = [round(i) for i in blueValues]
+            blueValues = [otRound(i) for i in blueValues]
         otherBlues = getAttrWithFallback(info, "postscriptOtherBlues")
         if isinstance(otherBlues, list):
-            otherBlues = [round(i) for i in otherBlues]
+            otherBlues = [otRound(i) for i in otherBlues]
         familyBlues = getAttrWithFallback(info, "postscriptFamilyBlues")
         if isinstance(familyBlues, list):
-            familyBlues = [round(i) for i in familyBlues]
+            familyBlues = [otRound(i) for i in familyBlues]
         familyOtherBlues = getAttrWithFallback(info, "postscriptFamilyOtherBlues")
         if isinstance(familyOtherBlues, list):
-            familyOtherBlues = [round(i) for i in familyOtherBlues]
+            familyOtherBlues = [otRound(i) for i in familyOtherBlues]
         stemSnapH = getAttrWithFallback(info, "postscriptStemSnapH")
         if isinstance(stemSnapH, list):
-            stemSnapH = [round(i) for i in stemSnapH]
+            stemSnapH = [otRound(i) for i in stemSnapH]
         stemSnapV = getAttrWithFallback(info, "postscriptStemSnapV")
         if isinstance(stemSnapV, list):
-            stemSnapV = [round(i) for i in stemSnapV]
+            stemSnapV = [otRound(i) for i in stemSnapV]
         # only write the blues data if some blues are defined.
         if any((blueValues, otherBlues, familyBlues, familyOtherBlues)):
             private.rawDict["BlueFuzz"] = blueFuzz
@@ -1048,6 +1105,7 @@ class OutlineTTFCompiler(BaseOutlineCompiler):
         glyf.glyphs = {}
         glyf.glyphOrder = self.glyphOrder
 
+        hmtx = self.otf["hmtx"]
         allGlyphs = self.allGlyphs
         for name in self.glyphOrder:
             glyph = allGlyphs[name]
@@ -1060,27 +1118,33 @@ class OutlineTTFCompiler(BaseOutlineCompiler):
             else:
                 ttGlyph = pen.glyph()
                 if ttGlyph.isComposite() and self.autoUseMyMetrics:
-                    self.autoUseMyMetrics(ttGlyph, glyph.width, allGlyphs)
+                    self.autoUseMyMetrics(ttGlyph, name, hmtx)
             glyf[name] = ttGlyph
 
     @staticmethod
-    def autoUseMyMetrics(ttGlyph, width, glyphSet):
+    def autoUseMyMetrics(ttGlyph, glyphName, hmtx):
         """ Set the "USE_MY_METRICS" flag on the first component having the
         same advance width as the composite glyph, no transform and no
         horizontal shift (but allow it to shift vertically).
         This forces the composite glyph to use the possibly hinted horizontal
         metrics of the sub-glyph, instead of those from the "hmtx" table.
         """
+        width = hmtx[glyphName][0]
         for component in ttGlyph.components:
             try:
                 baseName, transform = component.getComponentInfo()
             except AttributeError:
                 # component uses '{first,second}Pt' instead of 'x' and 'y'
                 continue
-            if (glyphSet[baseName].width == width and
-                    transform[:-1] == (1, 0, 0, 1, 0)):
-                component.flags |= USE_MY_METRICS
-                break
+            try:
+                baseMetrics = hmtx[baseName]
+            except KeyError:
+                continue  # ignore missing components
+            else:
+                if (baseMetrics[0] == width and
+                        transform[:-1] == (1, 0, 0, 1, 0)):
+                    component.flags |= USE_MY_METRICS
+                    break
 
 
 class StubGlyph(object):
@@ -1098,6 +1162,7 @@ class StubGlyph(object):
         self.descender = descender
         self.unicodes = unicodes
         self.components = []
+        self.anchors = []
         if unicodes:
             self.unicode = unicodes[0]
         else:
@@ -1110,28 +1175,16 @@ class StubGlyph(object):
             return 1
         return 0
 
-    def _get_leftMargin(self):
-        if self.bounds is None:
-            return 0
-        return self.bounds[0]
-
-    leftMargin = property(_get_leftMargin)
-
-    def _get_rightMargin(self):
-        bounds = self.bounds
-        if bounds is None:
-            return 0
-        xMin, yMin, xMax, yMax = bounds
-        return self.width - bounds[2]
-
-    rightMargin = property(_get_rightMargin)
+    @property
+    def height(self):
+        return self.ascender - self.descender
 
     def draw(self, pen):
         pass
 
     def _drawDefaultNotdef(self, pen):
-        width = round(self.unitsPerEm * 0.5)
-        stroke = round(self.unitsPerEm * 0.05)
+        width = otRound(self.unitsPerEm * 0.5)
+        stroke = otRound(self.unitsPerEm * 0.05)
         ascender = self.ascender
         descender = self.descender
         xMin = stroke
